/* Copyright (c) 2025, NVIDIA CORPORATION & AFFILIATES.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#pragma once

#include "xla/ffi/api/c_api.h"
#include "xla/ffi/api/ffi.h"


XLA_FFI_DECLARE_HANDLER_SYMBOL(OperatorActionHandler);
XLA_FFI_DECLARE_HANDLER_SYMBOL(OperatorActionBackwardDiffHandler);
